/*
 * Enhanced Speedstep & PowerNow Daemon for NetBSD & DragonFly BSD, Code (c) 
 * 2004-2007 by Ove Soerensen, Portions (c) 2006 Johannes Hofmann, 2007
 * Stephen M. Rumble
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met: 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer. 2.
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS 'AS IS' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE   FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/param.h>
#include <sys/sysctl.h>
#if defined(__DragonFly__)
 #include <kinfo.h>
#else
 #include <sys/sched.h>
 #include <util.h>
#endif
#include <errno.h>
#include <signal.h>

#define ESTD_VERSION "Release 6"
#define BATTERY 0
#define SMOOTH 1
#define AGGRESSIVE 2
#define DEF_POLL 500000
#define MIN_POLL 10000
#define DEF_HIGH 80
#define DEF_LOW 40
#define TECH_UNKNOWN 0
#define TECH_EST 1
#define TECH_POWERNOW 2
#define TECH_MAX 2

/* this is ugly, but... <shrug> */
#define MAX_FREQS 32
#define SYSCTLBUF 255

extern char    *optarg;
extern int      optind;

#if defined(__DragonFly__)
 #define useconds_t unsigned int
#endif

/* command-line options */
int             daemonize = 0;
int             verbose = 0;
int             nicemod = 0;
int             strategy = SMOOTH;
useconds_t      poll = DEF_POLL;
int             high = DEF_HIGH;
int             low = DEF_LOW;
useconds_t      lowgrace = 0;
int             minmhz = -1;
int             maxmhz = -1;
int             listfreq = 0;
int             tech = TECH_UNKNOWN;
int             use_clockmod = 0;
int             clockmod_min = -1;
int             clockmod_max = -1;

#if defined(__DragonFly__)
 static struct kinfo_cputime cp_time;
 static struct kinfo_cputime cp_old;
 static struct kinfo_cputime cp_diff;
#else
 static int cpumib[2] = {CTL_KERN, KERN_CP_TIME};
 static u_int64_t cp_time[CPUSTATES];
 static u_int64_t cp_old[CPUSTATES];
 static u_int64_t cp_diff[CPUSTATES];
 static size_t cp_time_size = sizeof(cp_time[0]) * CPUSTATES;
#endif

static char	*techdesc[3] = {"Unknown",
				"Enhanced SpeedStep",
				"PowerNow"
				};
static char	*freqctl[3] = {	"",	
				"machdep.est.frequency.available",
				"machdep.powernow.frequency.available"
				};
static char	*setctl[3] = {	"",
				"machdep.est.frequency.target",
				"machdep.powernow.frequency.target"
				};

void
usage()
{
	printf("usage: estd [-d] [-o] [-n] [-C] [-E] [-P] [-a] [-s] [-b] [-p poll interval in us] [-g grace period] [-l low watermark percentage] [-h high watermark percentage] [-m minimum MHz] [-M maximum MHz]\n");
	printf("       estd -v\n");
	printf("       estd -f\n");
	exit(1);
}

void
version()
{
	printf("estd ");
	printf(ESTD_VERSION);
	printf(" Copyright (c) 2004-2007 Ove Soerensen\n");
	printf("Details at http://www.elektro-eel.org - Contact via ove@elektro-eel.org\n");
	exit(0);
}

/* returns cpu-usage in percent, mean over the sleep-interval or -1 if an error occured */
#if defined(__DragonFly__)
int
get_cpuusage()
{
	int 		i;
	u_int64_t	total_time = 0;

	cp_old = cp_time;
	if (kinfo_get_sched_cputime(&cp_time)) {
		fprintf(stderr, "estd: Cannot get CPU status\n");
		exit(1);
	}

	cp_diff.cp_user = cp_time.cp_user - cp_old.cp_user;
	total_time += cp_diff.cp_user;
	cp_diff.cp_nice = cp_time.cp_nice - cp_old.cp_nice;
	total_time += cp_diff.cp_nice;
	cp_diff.cp_sys = cp_time.cp_sys - cp_old.cp_sys;
	total_time += cp_diff.cp_sys;
	cp_diff.cp_intr = cp_time.cp_intr - cp_old.cp_intr;
	total_time += cp_diff.cp_intr;
	cp_diff.cp_idle = cp_time.cp_idle - cp_old.cp_idle;
	total_time += cp_diff.cp_idle;

	/* we've probably been interrupted by a signal... */
	if (total_time < 1) return -1;

	return (100 - ((cp_diff.cp_idle + (cp_diff.cp_nice * nicemod)) * 100) / total_time);
}
#else
int
get_cpuusage()
{
	int             i;
	u_int64_t       total_time = 0;
	memcpy(&cp_old[0], &cp_time[0], cp_time_size);
	if (sysctl(cpumib, 2, &cp_time, &cp_time_size, NULL, 0) < 0) {
		fprintf(stderr, "estd: Cannot get CPU status\n");
		exit(1);
	}
	for (i = 0; i < CPUSTATES; i++) {
		cp_diff[i] = cp_time[i] - cp_old[i];
		total_time += cp_diff[i];
	}

	/* we've probably been interrupted by a signal... */
	if (total_time < 1) return -1; 

	return (100 - ((cp_diff[CP_IDLE] + (cp_diff[CP_NICE] * nicemod)) * 100) / total_time);
}
#endif

/* sets the cpu frequency */
void
set_freq(int freq)
{
	if ((!daemonize) && (verbose))
		printf("%i MHz\n", freq);
	if (sysctlbyname(setctl[tech], NULL, NULL, &freq, sizeof(freq)) < 0) {
		fprintf(stderr, "estd: Cannot set CPU frequency (maybe you aren't root?)\n");
		exit(1);
	}
}

void
set_clockmod(int level)
{
#if !defined(__DragonFly__)
	if (!use_clockmod || level == -1)
		return;

	if ((!daemonize) && (verbose))
		printf("clockmod level: %i\n", level);
	if (sysctlbyname("machdep.clockmod.target", NULL, NULL, &level,
	    sizeof(level)) < 0) {
		fprintf(stderr, "estd: Cannot set clockmod level (maybe you aren't root?)\n");
		exit(1);
	}
#endif
}

/* need this callback for sorting the frequency-list */
int
freqcmp(const void *x, const void *y)
{
	return *((int *) x) - *((int *) y);
}

/* clean up the pidfile and clockmod on exit */
void
sighandler(int sig)
{
	set_clockmod(clockmod_max);
	exit(0);
}

/* switch strategy on SIGUSR{1,2} */
void
sigusrhandler(int sig)
{
	switch (sig) {
		case SIGUSR1:
				if (strategy>BATTERY) strategy--;
				break;
		case SIGUSR2:
				if (strategy<AGGRESSIVE) strategy++;
				break;
	}
}

int
main(int argc, char *argv[])
{
	int             ch;
	int             i;
	char            frequencies[SYSCTLBUF];	/* XXX Ugly */
	char           *fp;
	size_t          freqsize = SYSCTLBUF;
	int             freqtab[MAX_FREQS];
	int             nfreqs = 0;
	int             minidx = 0;
	int             maxidx = 0;
	int             curcpu = 0;
	int             curfreq = 0;
	int	            curstrat = strategy;
	useconds_t		lowtime = 0;
	FILE           *fexists;

	/* get command-line options */
	while ((ch = getopt(argc, argv, "vfdonCEPasbp:h:l:g:m:M:")) != -1)
		switch (ch) {
		case 'v':
			version();
			/* NOTREACHED */
		case 'f':
			listfreq = 1;
			break;
		case 'd':
			daemonize = 1;
			break;
		case 'o':
			verbose = 1;
			break;
		case 'n':
			nicemod = 1;
			break;
		case 'C':
			#if !defined(__DragonFly__)
			 use_clockmod = 1;
			 break;
			#else
			 fprintf(stderr, "-C not available under DragonFly\n");
			 exit(1);
			#endif
		case 'E':
			tech = TECH_EST;
			break;
		case 'P':
			tech = TECH_POWERNOW;
		case 'a':
			strategy = AGGRESSIVE;
			break;
		case 's':
			strategy = SMOOTH;
			break;
		case 'b':
			strategy = BATTERY;
			break;
		case 'p':
			poll = atoi(optarg);
			break;
		case 'h':
			high = atoi(optarg);
			break;
		case 'l':
			low = atoi(optarg);
			break;
		case 'g':
			lowgrace = atoi(optarg);
			break;
		case 'm':
			minmhz = atoi(optarg);
			break;
		case 'M':
			maxmhz = atoi(optarg);
			break;
		default:
			usage();
			/* NOTREACHED */
		}

	/* try to guess cpu-scaling technology */
	if (tech == TECH_UNKNOWN) {
		for (tech = 1; tech <= TECH_MAX; tech++) {
			if (sysctlbyname(freqctl[tech], &frequencies, &freqsize, NULL, 0) >= 0) break;
		}
		if (tech > TECH_MAX) {
			fprintf(stderr, "estd: Cannot guess CPU-scaling technology. (maybe you are missing some kernel-option?)\n");
			exit(1);
		}
	}
	
	/* get supported frequencies... */
 	if (sysctlbyname(freqctl[tech], &frequencies, &freqsize, NULL, 0) < 0) {
		fprintf(stderr, "estd: Cannot get supported frequencies (maybe you forced the wrong CPU-scaling technology?)\n");
		exit(1);
	}
	fp = &frequencies[0];
	while ((nfreqs < MAX_FREQS) && ((freqtab[nfreqs++] = strtol(fp, &fp, 10)) != 0));
	nfreqs--;
	if (nfreqs <= 0) {
		fprintf(stderr, "estd: No supported frequencies found?! (please report this error)\n");
		exit(1);
	}
	/* ...and sort them in ascending order */
	qsort(&freqtab, nfreqs, sizeof(freqtab[0]), &freqcmp);

	if (listfreq) {
		printf("Supported frequencies (%s Mode):\n",techdesc[tech]);
		for (i = 0; i < nfreqs; i++) {
			printf("%i MHz\n", freqtab[i]);
		}
		exit(0);
	}
	
	/* some sanity checks */
	if ((high <= low) || (low < 0) || (low > 100) || (high < 0) || (high > 100)) {
		fprintf(stderr, "estd: Invalid high/low watermark combination\n");
		exit(1);
	}
	if (minmhz < 0)
		minmhz = freqtab[0];
	if (maxmhz < 0)
		maxmhz = freqtab[nfreqs - 1];
	if (minmhz > maxmhz) {
		fprintf(stderr, "estd: Invalid minimum/maximum MHz combination\n");
		exit(1);
	}
	while ((minidx < nfreqs) && (freqtab[minidx] < minmhz))
		minidx++;
	if (minidx >= nfreqs) {
		fprintf(stderr, "estd: Minimum Frequency is too high\n");
		exit(1);
	}
	maxidx = nfreqs - 1;
	while ((maxidx > -1) && (freqtab[maxidx] > maxmhz))
		maxidx--;
	if (maxidx < 0) {
		fprintf(stderr, "estd: Maximum Frequency is too low\n");
		exit(1);
	}
	if (freqtab[minidx] > freqtab[maxidx]) {
		fprintf(stderr, "estd: No supported frequency within given range found\n");
		exit(1);
	}
	if (poll < MIN_POLL) {
		fprintf(stderr, "estd: Poll interval is too low (minimum %i)\n", MIN_POLL);
		exit(1);
	}

	#if !defined(__DragonFly__)
	{
		char   *lastfp = NULL;
		char	clockmods[SYSCTLBUF];
		size_t	len = sizeof(clockmods);

		if (sysctlbyname("machdep.clockmod.available", &clockmods, &len,
		    NULL, 0) >= 0) {
			fp = &clockmods[0];
			while (fp != lastfp) {
				lastfp = fp;
				i = strtol(fp, &fp, 10);
				if (i < clockmod_min || clockmod_min == -1)
					clockmod_min = i;
				if (i > clockmod_max || clockmod_max == -1)
					clockmod_max = i;
			}
		}
	}
	#endif

	if ((fexists = fopen("/var/run/estd.pid", "r")) != NULL) {
		fprintf(stderr, "estd: Pidfile /var/run/estd.pid exists, remove it if you are sure it shouldn't be there (maybe another instance of estd is already running?)\n");
		fclose(fexists);
		exit(1);
	}
	
	/* all ok, here we go */
	if (daemonize) {
		if (fork()) {
			printf("estd: Forked\n");
			exit(0);
		}
	} else {
		printf("estd: Not detaching from terminal\n");
	}

	if (pidfile(NULL)) {
		fprintf(stderr, "estd: Cannot write pidfile (maybe you aren't root?)\n");
		exit(1);
	}
	
	/* init some vars and set inital frequency */
	signal(SIGHUP, SIG_IGN);
	signal(SIGINT, &sighandler);
	signal(SIGPIPE, &sighandler);
	signal(SIGTERM, &sighandler);
	signal(SIGUSR1, &sigusrhandler);
	signal(SIGUSR2, &sigusrhandler);
	curfreq = minidx;
	set_freq(freqtab[curfreq]);
	set_clockmod(clockmod_min);

	#if !defined(__DragonFly__)
	 for (i = 0; i < CPUSTATES; i++) {
	 	cp_time[i] = 0;
	 	cp_old[i] = 0;
		cp_diff[i] = 0;
	 }
	#endif

	/* the big processing loop, we will only exit via signal */
	while (1) {
		curcpu = get_cpuusage();
		if (curcpu != -1) {
			/* strategy can change anytime (SIGUSR) */ 
			curstrat = strategy;
			if ((curfreq > minidx) && (curcpu < low)) {
				if (lowtime < lowgrace)
					lowtime += poll;

				if (lowtime >= lowgrace) {
					if (curstrat == BATTERY)
						curfreq = minidx;
					else
						curfreq--;
					set_freq(freqtab[curfreq]);
					if (curfreq == minidx)
						set_clockmod(clockmod_min);
				}
			} else {
				lowtime = 0;

				if ((curfreq < maxidx) && (curcpu > high)) {
					if (curstrat == AGGRESSIVE)
						curfreq = maxidx;
					else
						curfreq++;
					set_freq(freqtab[curfreq]);
					if (curfreq != minidx)
						set_clockmod(clockmod_max);
				}
			}
		}
		usleep(poll);
	}

	return 0;
}

